/*
 * Copyright (c) 2015 Roc Streaming authors
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

//! @file roc_core/target_posix/roc_core/stddefs.h
//! @brief Commonly used types and functions.

#ifndef ROC_CORE_STDDEFS_H_
#define ROC_CORE_STDDEFS_H_

#include <algorithm>
#include <cmath>
#include <ctime>
#include <new>

#include <ctype.h>
#include <inttypes.h>
#include <limits.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#endif // ROC_CORE_STDDEFS_H_
