/*
 * Copyright (c) 2019 Roc Streaming authors
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include "roc_audio/iresampler.h"

namespace roc {
namespace audio {

IResampler::IResampler(core::IArena& arena)
    : core::RefCounted<IResampler, core::ArenaAllocation>(arena) {
}

IResampler::~IResampler() {
}

} // namespace audio
} // namespace roc
